/*
 * Copyright 2000-2001 by Jean-Claude Christophe <jch@oleane.net>
 *                                               <jch@bsdfr.org>
 *
 * Home page:  http://www.bsdfr.org/projects/tac_bsd/
 *
 * login.c  This file contains functions used when the user is authenticated
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ttyent.h>
#include <grp.h>      /* initgroups(3) */
#include <paths.h>
#include <syslog.h>
#include <utmp.h>
#include <pwd.h>
#include <strings.h>
#include <string.h>
#include <utmp.h>     /* utmp struct */
#include <time.h>     /* time(3) */

#ifdef HAVE_UTIL_H
#include <util.h>     /* login(3) */
#endif
#ifdef HAVE_LIBUTIL_H
#include <libutil.h>  /* login(3) */
#endif

#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif

#ifdef HAVE_SHADOW_H
#include <shadow.h>
#endif

#include "tac_bsd.h"
#include "conf.h"

#define LOCAL 1
#define REMOTE 2

/* in connect.c */
extern int connect_to_srv(const char *);

/* in packet.c */
extern int send_authen(int, char*, char *, char *);
extern int send_auth_cont(int, char *, int, char *);
extern int read_authen(int, char *);
extern int read_result(int, char *);

/* in main.c */
extern void close_quit(void);
extern void display_file(char *);

uid_t uid_default;     /* uid of our user */
uid_t gid_default;     /* gid of our user */

struct passwd *pwd;    /* if user is local */

#ifdef HAVE_SHADOW_H
struct spwd *spwd;     /* using shadow password */
#endif

extern int done, try;  /* in main.c */
int login_status;      /* local or remote. used by login() in order to know how to proceed */

char *
stypeof(ttyid)
    char *ttyid;
{
    struct ttyent *t;

    return (ttyid && (t = getttynam(ttyid)) ? t->ty_type : "unknown");
}

/* 
 * If we arrived here, all seems to be good.
 */
void
login_in()
{
    struct utmp utmp;
    char *home = (login_status == LOCAL ? pwd->pw_dir : directory);
    uid_t myuid = (login_status == LOCAL ? pwd->pw_uid : uid_default);
    gid_t mygid = (login_status == LOCAL ? pwd->pw_gid : gid_default);
    char *myshell = (login_status == LOCAL ? pwd->pw_shell : shell);

    /* fill the utmp struct */
    memset(&utmp, 0, sizeof(utmp));
    time(&utmp.ut_time);
    strncpy(utmp.ut_name, username, sizeof(utmp.ut_name));
    strncpy(utmp.ut_line, port + 5, sizeof(utmp.ut_line));
    if (hostname)
	strncpy(utmp.ut_host, hostname, sizeof(utmp.ut_host));
    login(&utmp);
    
    /* tty is our now */
    chown(port, myuid, mygid);

    setegid(mygid);
    setgid(mygid);
    initgroups(username, mygid);
    setuid(myuid);
    seteuid(myuid);

#ifdef HAVE_LOGIN_FBTAT
    login_fbtab(port + 5, myuid, mygid);
#endif
	
    term = stypeof(port + 5);     /* +5 for remove the "/dev/" */

    setenv("TERM", term, 0);
    setenv("LOGNAME", username, 1);
    setenv("USER", username, 1);
    setenv("PATH", _PATH_DEFPATH, 1);

    if (chdir(home) < 0) {
	printf("No home directory %s\n", home);
	setenv("HOME", "/", 1);
    } else 
	setenv("HOME", home, 1);
    
    setenv("SHELL", myshell, 1);

    /* 
     * It's time to add some message if the user want them.
     * if user have a .hushlogin in his home directory, we display nothing.
     * else, we display first "motdpass" (if it exists) and "/etc/motd" (if it exists too)
     */

    /* first, we check if ~/.hushlogin exists */
    if (access(HUSHLOGIN, F_OK) < 0) {
	/* it doesn't exist, we display the file(s) */
	if (motdpass)
	    display_file(motdpass);
	display_file(MOTD_FILE); /* display_file will check if the file exist or not */
    }

#ifdef HAVE_SETLOGIN
    setlogin(username);
#endif

/*    tbuf[0] = '-';
    strncpy(tbuf + 1, (p = strrchr(myshell, '/')) ? p + 1 : myshell, sizeof tbuf - 1 - 1);
    tbuf[sizeof tbuf - 1] = '\0';*/

    execlp(myshell, myshell, 0); /* our work stop here. */
}

/* 
 * returns 1 if the user is in the local password DB
 * -1 in other case.
 */
int
local_user()
{
    if (((pwd = getpwnam(username))) || ((!serv) && (serv_count > 1)))
	return(1);

    return(-1);
}

int
login_local()
{
#ifndef HAVE_SHADOW_H
    char *salt = (pwd = getpwnam(username)) ? pwd->pw_passwd : NULL;
#else
    char *salt = (spwd = getspnam(username)) ? spwd->sp_pwdp : NULL;
#endif

    /* ALWAYS ask for a password */
    password = getpass("Password: ");

    if (!salt) {
	authen_failed();
	sprintf(tac_err, "inexistent login");
	return(-1);
    }

#ifdef HAVE_SHADOW_H
    pwd = getpwnam(username);
#endif

    /* check if uid 0 */
    if (!pwd->pw_uid && !permit_uid0) {
	authen_failed();
	sprintf(tac_err, "try to log with uid 0");
	return(-1);
    }

    /* check the password */
#ifndef HAVE_SHADOW_H
    if (strcmp((char *)crypt(password, salt), pwd->pw_passwd)) {
#else
    if (strcmp((char *)crypt(password, salt), spwd->sp_pwdp)) {
#endif
	authen_failed();
	sprintf(tac_err, "bad password");
	return(-1);
    }
    memset(password, 0, strlen(password));
    
    done = 1; /* local authentication correct */
    login_status = LOCAL;

    endpwent();
    return(1);
}

int
login_remote()
{
    int srv_sd;   /* socket used by the server */
    int result = FALSE;
    unsigned char count = 0; /* default: no '@' in username, only 1 serv defined */
    
    if (!serv && serv_count > 1) {
	/* 
	 * Account is invalid, however we ask for a password
	 */
	password = getpass("Password: ");
        return(-1);
    } else if (serv_count < 1)
        serv++;
    
    if (serv_count > 1) {
	serv++;
        /* searching the good struct in tac_serv */
        for (count = 0; strlen(tac_serv[count].name); count++) {
            if (!strncmp(serv, tac_serv[count].name, strlen(serv))) {
                result = TRUE;
                break; /* keep the count value */
            }
        }
        /* did we find the server we searched for authentication ? */
        if (!result)
            return(-1); /* no */
    }
    
    /* First step: connection and START request */
    if (tac_serv[count].addr != '\0') {
	srv_sd = connect_to_srv(tac_serv[count].addr);
    } else
	return(-1);

    if (srv_sd < 0)
	return (-1);

    if (send_authen(srv_sd, strtok(username, "@"), port, tac_serv[count].key ? tac_serv[count].key : NULL) < 0) {
	if (slog) 
	    fprintf(stderr, "Error in first step: check into syslog for details\n");
	close_quit();
    }

    /* Second step: wait the response, we're expecting a GETPASS */
    if (read_authen(srv_sd, tac_serv[count].key ? tac_serv[count].key : NULL) < 0) {
	if (slog) 
	    fprintf(stderr, "Error in second step: check into syslog for details\n");
	close_quit();
    }
    /* Third step: send the CONTINUE packet */
    if (send_auth_cont(srv_sd, password, strlen(password), tac_serv[count].key ? tac_serv[count].key : NULL) < 0) {
	if (slog) 
	    fprintf(stderr, "Error in third step: check into syslog for details\n");
	close_quit();
    }

    /* Fourth step: get the fatal result: PASS or FAIL */
    result = read_result(srv_sd, tac_serv[count].key ? tac_serv[count].key : NULL);

    if (result != -1) {
	done = 1; /* done */
	login_status = REMOTE;
    }
    shutdown(srv_sd, SHUT_RDWR); close(srv_sd);

    return(result); /* -1: failed, 0: ok */
}
